"""
Test module for runpandas acessors
"""

import os

import pytest
import pandas as pd
from runpandas import reader
from runpandas.exceptions import RequiredColumnError
from runpandas.types import columns

pytestmark = pytest.mark.stable


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture
def pandas_gpx_activity(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    return reader._read_file(gpx_file, to_df=True)


@pytest.fixture
def runpandas_gpx_activity(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    return reader._read_file(gpx_file, to_df=False)


@pytest.fixture
def pandas_tcx_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    return reader._read_file(tcx_file, to_df=True)


@pytest.fixture
def runpandas_tcx_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "stopped_example.tcx")
    return reader._read_file(tcx_file, to_df=False)


@pytest.fixture
def runpandas_tcx_basic_activity(dirpath):
    tcx_file = os.path.join(dirpath, "tcx", "basic.tcx")
    return reader._read_file(tcx_file, to_df=False)


def test_metrics_validate(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")

    with pytest.raises(AssertionError):
        activity_gpx = reader._read_file(gpx_file, to_df=True)
        activity_gpx.compute.speed()

    activity_gpx = reader._read_file(gpx_file, to_df=False)

    # copy of dataframe for testing the index error
    frame_without_index = activity_gpx.copy()
    frame_without_index = frame_without_index.reset_index()
    with pytest.raises(AttributeError):
        frame_without_index.compute.speed()

    # test the decorator special_column
    activity_without_required_column = activity_gpx.drop(["lat"], axis=1)
    with pytest.raises(RequiredColumnError):
        activity_without_required_column.compute.distance()

    activity_without_required_column = activity_gpx.drop(["alt"], axis=1)
    with pytest.raises(RequiredColumnError):
        activity_without_required_column.compute.distance(
            correct_distance=True, to_special_column=False
        )


test_distance_pos_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "distpos", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "distpos", 2, 5.11816186976935),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "distpos", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "distpos", 2, 5.11816186976935),
]


@pytest.mark.parametrize("activity,column,index,expected", test_distance_pos_data)
def test_metrics_distance(activity, column, index, expected):
    activity["distpos"] = activity.compute.distance(to_special_column=False)
    assert activity[column].iloc[index] == expected


test_correct_distance_pos_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "distpos", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "distpos", 2, 5.11816186976935),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "distpos", -1, 5.093444269601765),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "distpos", 2, 5.118301239044248),
]


@pytest.mark.parametrize(
    "activity,column,index,expected", test_correct_distance_pos_data
)
def test_metrics_corrected_distance(activity, column, index, expected):
    activity["distpos"] = activity.compute.distance(
        correct_distance=True, to_special_column=False
    )
    assert activity[column].iloc[index] == expected


test_full_distpos_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "distpos", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "distpos", 2, 5.11816186976935),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "distpos", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "distpos", 2, 5.11816186976935),
]


@pytest.mark.parametrize("activity,column,index,expected", test_full_distpos_data)
def test_metrics_distance_per_position(activity, column, index, expected):
    activity["distpos"] = activity.compute.distance(to_special_column=True)
    assert isinstance(activity["distpos"], columns.DistancePerPosition)
    assert activity[column].iloc[index] == expected


test_speed_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "speed", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "speed", 2, 5.11816186976935),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "speed", -1, 5.093437453100809),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "speed", 2, 5.11816186976935),
]


def test_speed_validate(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")

    with pytest.raises(RequiredColumnError):
        activity_gpx = reader._read_file(gpx_file, to_df=False)
        assert "dist_pos" not in activity_gpx.columns
        activity_gpx.compute.speed(from_distances=True)

    with pytest.raises(RequiredColumnError):
        activity_gpx = reader._read_file(gpx_file, to_df=False)
        assert "speed" not in activity_gpx.columns
        activity_gpx.compute.speed(from_distances=False)

    with pytest.raises(RequiredColumnError):
        activity_gpx = reader._read_file(gpx_file, to_df=False)
        assert "speed" not in activity_gpx.columns
        activity_gpx.compute.speed(from_distances=False)


def test_vam_validate(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_without_required_column = activity_gpx.drop(["alt"], axis=1)
    assert "alt" not in activity_without_required_column.columns
    with pytest.raises(RequiredColumnError):
        activity_without_required_column.compute.vertical_speed()


def test_gradient_validate(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")
    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_without_required_column = activity_gpx.drop(["alt"], axis=1)
    assert "alt" not in activity_without_required_column.columns
    with pytest.raises(RequiredColumnError):
        activity_without_required_column.compute.gradient()

    activity_gpx = reader._read_file(gpx_file, to_df=False)
    activity_gpx["distpos"] = activity_gpx.compute.distance()
    assert "dist" not in activity_gpx.columns
    with pytest.raises(RequiredColumnError):
        activity_gpx.compute.gradient()


test_speed_gpx_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "speed", -1, 2.5467187265504045),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "speed", 2, 2.559080934884675),
]


@pytest.mark.parametrize("activity,column,index,expected", test_speed_gpx_data)
def test_metrics_from_distances_speed(activity, column, index, expected):
    activity["distpos"] = activity.compute.distance(correct_distance=True)
    activity["speed"] = activity.compute.speed(from_distances=True)
    assert activity[column].iloc[index] == expected


test_speed_tcx_data = [
    (pytest.lazy_fixture("runpandas_tcx_activity"), "speed", -1, 2.36077),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "speed", 3, 2.222973),
]


@pytest.mark.parametrize("activity,column,index,expected", test_speed_tcx_data)
def test_metrics_from_recording_speed(activity, column, index, expected):
    activity["speed"] = activity.compute.speed(from_distances=False)
    assert activity[column].iloc[index] == expected


test_vam_gpx_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "vam", -4, 0.02499999999999991),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "vam", -5, 0.0),
]


@pytest.mark.parametrize("activity,column,index,expected", test_vam_gpx_data)
def test_metrics_gpx_vam(activity, column, index, expected):
    activity["vam"] = activity.compute.vertical_speed()
    assert activity[column].iloc[index] == expected


test_vam_tcx_data = [
    (pytest.lazy_fixture("runpandas_tcx_activity"), "vam", -1, -0.004166500000000073),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "vam", 3, 0.018022000000000205),
]


@pytest.mark.parametrize("activity,column,index,expected", test_vam_tcx_data)
def test_metrics_tcx_vam(activity, column, index, expected):
    activity["vam"] = activity.compute.vertical_speed()
    assert activity[column].iloc[index] == pytest.approx(expected, 0.01)


test_gradient_gpx_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "grad", -4, 0.006652869371284877),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "grad", -5, 0.0),
]


@pytest.mark.parametrize("activity,column,index,expected", test_gradient_gpx_data)
def test_metrics_gpx_gradient(activity, column, index, expected):
    activity["distpos"] = activity.compute.distance()
    activity["dist"] = activity["distpos"].to_distance()
    activity["grad"] = activity.compute.gradient()
    assert activity[column].iloc[index] == expected


test_gradient_tcx_data = [
    (pytest.lazy_fixture("runpandas_tcx_activity"), "grad", -1, -0.0013069369604063105),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "grad", 3, 0.005967282919472209),
]


@pytest.mark.parametrize("activity,column,index,expected", test_gradient_tcx_data)
def test_metrics_tcx_gradient(activity, column, index, expected):
    activity["grad"] = activity.compute.gradient()
    assert activity[column].iloc[index] == pytest.approx(expected, 0.01)


def test_pace_validate(dirpath):
    gpx_file = os.path.join(dirpath, "gpx", "stopped_example.gpx")

    with pytest.raises(RequiredColumnError):
        activity_gpx = reader._read_file(gpx_file, to_df=False)
        assert "speed" not in activity_gpx.columns
        activity_gpx.compute.pace()


test_pace_tcx_data = [
    (
        pytest.lazy_fixture("runpandas_tcx_activity"),
        "pace",
        -1,
        pd.Timedelta("0 days 00:00:00.423590"),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_activity"),
        "pace",
        3,
        pd.Timedelta("0 days 00:00:00.449848"),
    ),
]


@pytest.mark.parametrize("activity,column,index,expected", test_pace_tcx_data)
def test_metrics_tcx_pace(activity, column, index, expected):
    activity["speed"] = activity.compute.speed(from_distances=False)
    activity["pace"] = activity.compute.pace()
    assert activity[column].iloc[index].total_seconds() == pytest.approx(
        expected.total_seconds(), 0.1
    )


test_pace_gpx_data = [
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "pace",
        -1,
        pd.Timedelta("0 days 00:00:00.392662"),
    ),
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "pace",
        2,
        pd.Timedelta("0 days 00:00:00.390765"),
    ),
]


@pytest.mark.parametrize("activity,column,index,expected", test_pace_gpx_data)
def test_metrics_gpx_pace(activity, column, index, expected):
    activity["distpos"] = activity.compute.distance(correct_distance=True)
    activity["speed"] = activity.compute.speed(from_distances=True)
    activity["pace"] = activity.compute.pace()
    assert activity[column].iloc[index].total_seconds() == pytest.approx(
        expected.total_seconds()
    )


test_hr_zone_gpx_data = [
    (pytest.lazy_fixture("runpandas_gpx_activity"), "hr_zone", -1, "Z4"),
    (pytest.lazy_fixture("runpandas_gpx_activity"), "hr_zone", 2, "Z5"),
]


@pytest.mark.parametrize("activity,column,index,expected", test_hr_zone_gpx_data)
def test_metrics_gpx_hr_zone(activity, column, index, expected):
    activity["hr_zone"] = activity.compute.heart_zone(
        labels=["Rest", "Z1", "Z2", "Z3", "Z4", "Z5"],
        bins=[0, 92, 110, 129, 147, 166, 184],
    )
    assert activity[column].iloc[index] == expected


test_hr_zone_tcx_data = [
    (pytest.lazy_fixture("runpandas_tcx_activity"), "hr_zone", -1, "Z4"),
    (pytest.lazy_fixture("runpandas_tcx_activity"), "hr_zone", 5, "Z5"),
]


@pytest.mark.parametrize("activity,column,index,expected", test_hr_zone_tcx_data)
def test_metrics_tcx_hr_zone(activity, column, index, expected):
    activity["hr_zone"] = activity.compute.heart_zone(
        labels=["Rest", "Z1", "Z2", "Z3", "Z4", "Z5"],
        bins=[0, 92, 110, 129, 147, 166, 184],
    )
    assert activity[column].iloc[index] == expected


test_time_in_zone_tcx_data = [
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z0",
        pd.Timedelta(seconds=216, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z1",
        pd.Timedelta(seconds=4, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z2",
        pd.Timedelta(seconds=4, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z3",
        pd.Timedelta(seconds=964, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z4",
        pd.Timedelta(seconds=682, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z5",
        pd.Timedelta(seconds=121, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_tcx_basic_activity"),
        "time_in_zone",
        "Z6",
        pd.Timedelta(seconds=0, microseconds=0),
    ),
]


@pytest.mark.parametrize("activity,column,index,expected", test_time_in_zone_tcx_data)
def test_metrics_tcx_time_in_zone(activity, column, index, expected):
    time_in_zone = activity.compute.time_in_zone(
        labels=["Z0", "Z1", "Z2", "Z3", "Z4", "Z5", "Z6"],
        bins=[0, 99, 129, 149, 169, 184, 199, 220],
    )
    assert activity.ellapsed_time.total_seconds() == time_in_zone.sum().total_seconds()
    assert time_in_zone[index].total_seconds() == expected.total_seconds()


test_time_in_zone_gpx_data = [
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "time_in_zone",
        "Rest",
        pd.Timedelta(seconds=0, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "time_in_zone",
        "Z1",
        pd.Timedelta(seconds=112, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "time_in_zone",
        "Z2",
        pd.Timedelta(seconds=492, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "time_in_zone",
        "Z3",
        pd.Timedelta(seconds=795, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "time_in_zone",
        "Z4",
        pd.Timedelta(seconds=3691, microseconds=0),
    ),
    (
        pytest.lazy_fixture("runpandas_gpx_activity"),
        "time_in_zone",
        "Z5",
        pd.Timedelta(seconds=37, microseconds=0),
    ),
]


@pytest.mark.parametrize("activity,column,index,expected", test_time_in_zone_gpx_data)
def test_metrics_gpx_time_in_zone(activity, column, index, expected):
    time_in_zone = activity.compute.time_in_zone(
        labels=["Rest", "Z1", "Z2", "Z3", "Z4", "Z5"],
        bins=[0, 92, 110, 129, 147, 166, 184],
    )
    assert activity.ellapsed_time.total_seconds() == time_in_zone.sum().total_seconds()
    assert time_in_zone[index].total_seconds() == expected.total_seconds()



"""
Tool to get/refresh Strava Token.
"""
from stravalib.client import Client
import os
import json
import time
import webbrowser
from json.decoder import JSONDecodeError
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlsplit, parse_qs


def coalesce(iterable):
    return next((el for el in iterable if el is not None), None)


class HTTPResponder(HTTPServer):
    allow_reuse_address = True
    timeout = 60
    access_token = None

    def handle_timeout(self):
        self.server_close()


class HTTPServerHandler(BaseHTTPRequestHandler):

    """
    HTTP Server callbacks to handle OAuth redirects
    """

    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        if "code" in self.path:
            self.auth_code = parse_qs(urlsplit(self.path).query)["code"]
            self.wfile.write(
                bytes(
                    "<html><h1>You may now close this window." + "</h1></html>",
                    "utf-8",
                )
            )
            self.server.access_token = self.auth_code

    # Disable logging from the HTTP Server
    def log_message(self, *args):
        return


class StravaClient(Client):
    """The StravaClient object is a helper tool for handling
    the authentication process (i.e. authorization, token update, ...) against the Strava.

    Parameters
    ----------
        token_file: File, The Strava access token path where it will be kept or loaded, optional
        refresh_token: str, The Strava refresh token, optional
        client_secret: str, The strava client secret used for token refresh, optional
        client_id: int, The Strava client id used for token refresh, optional

    """

    def __init__(
        self,
        *args,
        token_file=None,
        refresh_token=None,
        client_secret=None,
        client_id=None,
        **kwargs
    ):
        super(self.__class__, self).__init__(*args, **kwargs)
        self.token_file = coalesce((token_file, os.getenv("STRAVA_TOKEN_FILE", None)))
        self.client_secret = coalesce(
            (client_secret, os.getenv("STRAVA_CLIENT_SECRET", None))
        )
        self.client_id = coalesce((client_id, os.getenv("STRAVA_CLIENT_ID", None)))

        if self.token_file is not None:
            self.get_token_from_file()

        if refresh_token is not None:
            self.refresh_token = refresh_token

    def set_token_from_dict(self, token):
        """
        It extracts the token components from a token dict response.
        """

        self.access_token = token["access_token"]
        self.refresh_token = token["refresh_token"]
        self.token_expires_at = token["expires_at"]

    def get_token_from_file(self):
        """
        Gets the token from a token_file and returns as dict.
        """
        if self.token_file is not None:
            with open(self.token_file, "r") as f:
                try:
                    token = json.load(f)
                except JSONDecodeError:
                    print(
                        "problems on reading %s file. It considers an empty file."
                        % self.token_file
                    )
                    token = None
        if token is not None:
            if type(token) == str:
                token = json.loads(token)
            self.set_token_from_dict(token)

    def save_token_to_file(self, token):
        """
        It saves the token to the file

        Parameters
        ----------
        token: File, the file where the token will be stored at.
        """
        if self.token_file is None:
            return None
        with open(self.token_file, "w") as f:
            f.write(json.dumps(token))

    def refresh(self):
        """
        It updates the access token if the token is expired.
        """
        if time.time() > self.token_expires_at:
            token = self.refresh_access_token(
                client_id=self.client_id,
                client_secret=self.client_secret,
                refresh_token=self.refresh_token,
            )
            if self.token_file is not None:
                with open(self.token_file, "w") as f:
                    f.write(json.dumps(token))

    def authenticate_web(self):
        """
        It handles the authentication web negotiation with Strava.

        Returns
        -------
        token_response: dict, the token response from Strava.

        """
        scope = [
            "read",
            "read_all",
            "profile:read_all",
            "activity:read",
            "activity:read_all",
        ]
        authorize_url = self.authorization_url(
            client_id=self.client_id,
            redirect_uri="http://localhost:5000/authorized",
            scope=scope,
        )
        webbrowser.open(authorize_url)
        httpServer = HTTPResponder(
            ("localhost", 5000),
            HTTPServerHandler,
        )
        httpServer.handle_request()

        # Exchange code for token:
        token_response = self.exchange_code_for_token(
            client_id=self.client_id,
            client_secret=self.client_secret,
            code=httpServer.access_token,
        )
        # Save it to file so we can use it until it expires.
        if self.token_file is not None:
            with open(self.token_file, "w+") as f:
                json.dump(token_response, f)

        return token_response



import numpy as np
from pandas import Series, to_timedelta
from runpandas._utils import series_property


class ColumnsRegistrator(type):
    """
    We keep a mapping of column used names to classes.
    """

    REGISTRY = {}

    def __new__(metacls, name, bases, namespace):
        new_cls = super().__new__(metacls, name, bases, namespace)
        # We register each concrete class
        if name != "MeasureSeries":
            metacls.REGISTRY[new_cls.colname] = new_cls

        return new_cls


class MeasureSeries(Series, metaclass=ColumnsRegistrator):
    _metadata = ["colname", "base_unit"]

    #
    # Implement pandas methods
    #

    @property
    def _constructor(self):
        return self.__class__

    @property
    def _constructor_expanddim(self):
        from runpandas.types import Activity

        return Activity

    def __init__(self, data, *args, **kwargs):
        super().__init__(data, *args, **kwargs)
        self._name = self.__class__.colname  # use *class* attribute

    def __finalize__(self, other, method=None, **kwargs):
        """Propagate metadata from other to self."""
        for name in self._metadata:
            object.__setattr__(self, name, getattr(other, name, None))
        return self


class Altitude(MeasureSeries):
    colname = "alt"
    base_unit = "m"

    @property
    def ascent(self):
        """
        Returns the ascent altitudes during the activity.
        """
        deltas = self.diff()
        return Altitude(np.where(deltas > 0, deltas, 0), index=self.index)

    @property
    def descent(self):
        """
        Returns the descent altitudes during the activity.
        """
        deltas = self.diff()
        return Altitude(np.where(deltas < 0, deltas, 0), index=self.index)

    @series_property
    def ft(self):
        """
        Returns the altitude from metres to feet
        """
        return self * 3.28084


class Cadence(MeasureSeries):
    colname = "cad"
    base_unit = "rpm"


class DistancePerPosition(MeasureSeries):
    colname = "distpos"
    base_unit = "m"

    def to_distance(self):
        """
        Returns the cummulative distance
        """
        return Distance._from_discrete(self)

    @series_property
    def km(self):
        """Returns the distance converted from metres to kilometres"""
        return self / 1000

    @series_property
    def miles(self):
        """Returns the distance converted from metres to miles"""
        return self / 1000 * 0.621371


class Distance(MeasureSeries):
    colname = "dist"
    base_unit = "m"

    @classmethod
    def _from_discrete(cls, data, *args, **kwargs):
        return cls(data.cumsum(), *args, **kwargs)

    @series_property
    def km(self):
        """Returns the distance converted from metres to kilometres"""
        return self / 1000

    @series_property
    def miles(self):
        """Returns the distance converted from metres to miles"""
        return self / 1000 * 0.621371


class HeartRate(MeasureSeries):
    colname = "hr"
    base_unit = "bpm"


class LonLat(MeasureSeries):
    colname = "lonlat"
    base_unit = "degrees"

    @classmethod
    def _from_semicircles_to_degrees(cls, data, *args, **kwargs):
        # https://github.com/kuperov/fit/blob/master/R/fit.R
        deg = (data * 180 / 2**31 + 180) % 360 - 180
        return cls(deg, *args, **kwargs)

    @series_property
    def radians(self):
        """Returns the coordinates from degrees to radians"""
        return np.radians(self)


class Longitude(LonLat):
    colname = "lon"


class Latitude(LonLat):
    colname = "lat"


class Pace(MeasureSeries):
    colname = "pace"
    base_unit = "sec/m"

    @series_property
    def min_per_km(self):
        """
        Returns the pace converted from sec/m to min/km
        """
        return self * 1000

    @series_property
    def min_per_mile(self):
        """
        Returns the pace converted from sec/m to min/mile
        """
        return self * 1000 / 1.61


class Power(MeasureSeries):
    colname = "pwr"
    base_unit = "watts"


class Speed(MeasureSeries):
    colname = "speed"
    base_unit = "m/s"

    def to_pace(self):
        pace = to_timedelta(1 / self, unit="s")
        return Pace(pace)

    @series_property
    def kph(self):
        """
        Returns the speed converted from m/s to km/h
        """
        return self * 60**2 / 1000

    @property
    def mph(self):
        """
        Returns the speed converted from m/s to miles/h
        """
        # self.kph is already a Series
        return self.kph / 1.61


class Temperature(MeasureSeries):
    colname = "temp"
    base_unit = "degrees_C"


class VAM(MeasureSeries):
    colname = "vam"
    base_unit = "m/s"


class Gradient(MeasureSeries):
    colname = "grad"
    base_unit = "fraction"

    _metadata = ["_rise", "_run"] + MeasureSeries._metadata

    def __init__(self, *args, rise=None, run=None, **kwargs):
        if rise is not None and run is not None:
            super().__init__(rise / run, *args, **kwargs)
            self._rise, self._run = rise.values, run.values
        else:
            super().__init__(*args, **kwargs)
            self._rise, self._run = None, None

    # support method to metadata setup
    def _set_attrs(self, **kwargs):
        for attr in self._metadata:
            if attr in kwargs:
                self.__setattr__(attr, kwargs.get(attr))

    @series_property
    def pct(self):
        """It converts the fraction to percent (%)"""
        return self * 100

    @series_property
    def radians(self):
        """It converts fraction to radians"""
        return np.arctan2(self._rise, self._run)

    @series_property
    def degrees(self):
        """It converts fraction to degrees"""
        return np.rad2deg(np.arctan2(self._rise, self._run))



"""
Test module for Strava API reader base module
"""

import os
import json
import time
import pytest
from pandas import DataFrame, Timedelta, Timestamp
from runpandas import read_strava
from runpandas import types
from runpandas import StravaClient
from stravalib.protocol import ApiV3
from stravalib.client import Client
from stravalib.model import Stream

pytestmark = pytest.mark.stable


class MockResponse:
    def __init__(self, json_file):
        with open(json_file) as json_handler:
            self.json_data = json.load(json_handler)

    def json(self):
        return self.json_data


def mock_get_activity_streams(streams_file):
    """
    @TODO: I needed to mock the behavior the `stravalib.client.get_activity_streams`,
    it isn't the best alternative for mock the request from strava by passing a json file.
    """

    stream_mock = MockResponse(streams_file).json()
    entities = {}
    for key, value in stream_mock.items():
        value["type"] = key
        stream = Stream.deserialize(value)
        entities[stream.type] = stream
    return entities


@pytest.fixture
def dirpath(datapath):
    return datapath("io", "data")


@pytest.fixture(scope="session")
def valid_token_file(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("token.json")


@pytest.fixture
def strava_client(valid_token_file):
    file_handler = open(valid_token_file, "w")
    file_handler.write(
        '{"access_token": "2334444",  "refresh_token": "235555", "expires_at": %d}'
        % (time.time() + 3600)
    )
    file_handler.close()

    client = StravaClient(
        client_id="STRAVA_ID",
        client_secret="STRAVA_CLIENT_SECRET",
        token_file=valid_token_file,
    )
    return client


@pytest.fixture
def strava_activity(dirpath, mocker, strava_client):
    activity_json = os.path.join(dirpath, "strava", "activity.json")
    streams_json = os.path.join(dirpath, "strava", "streams.json")

    mocker.patch.object(ApiV3, "get", return_value=MockResponse(activity_json).json())

    mocker.patch.object(
        Client,
        "get_activity_streams",
        return_value=mock_get_activity_streams(streams_json),
    )
    # we don't use access token here, since we will mock the stravalib json response
    activity = read_strava(
        activity_id=4437021783,
        access_token=None,
        refresh_token=None,
        client=strava_client,
        to_df=False,
    )
    return activity


@pytest.fixture
def strava_dataframe(dirpath, mocker, strava_client):
    activity_json = os.path.join(dirpath, "strava", "activity.json")
    streams_json = os.path.join(dirpath, "strava", "streams.json")

    mocker.patch.object(ApiV3, "get", return_value=MockResponse(activity_json).json())

    mocker.patch.object(
        Client,
        "get_activity_streams",
        return_value=mock_get_activity_streams(streams_json),
    )
    # we don't use access token here, since we will mock the stravalib json response
    activity = read_strava(
        activity_id=4437021783,
        access_token=None,
        client=strava_client,
        refresh_token=None,
        to_df=True,
    )
    return activity


def test_read_strava_basic_dataframe(dirpath, mocker, strava_client):
    activity_json = os.path.join(dirpath, "strava", "activity.json")
    streams_json = os.path.join(dirpath, "strava", "streams.json")

    mocker.patch.object(ApiV3, "get", return_value=MockResponse(activity_json).json())

    mocker.patch.object(
        Client,
        "get_activity_streams",
        return_value=mock_get_activity_streams(streams_json),
    )
    # we don't use access token here, since we will mock the stravalib json response
    activity = read_strava(
        activity_id=4437021783,
        access_token=None,
        refresh_token=None,
        client=strava_client,
        to_df=True,
    )
    assert isinstance(activity, DataFrame)
    included_data = set(
        [
            "latitude",
            "longitude",
            "altitude",
            "distance",
            "velocity_smooth",
            "heartrate",
            "cadence",
            "moving",
            "grade_smooth",
        ]
    )
    assert included_data <= set(activity.columns.to_list())
    assert activity.size == 15723


def test_read_strava_activity(dirpath, mocker, strava_client):
    activity_json = os.path.join(dirpath, "strava", "activity.json")
    streams_json = os.path.join(dirpath, "strava", "streams.json")

    mocker.patch.object(ApiV3, "get", return_value=MockResponse(activity_json).json())

    mocker.patch.object(
        Client,
        "get_activity_streams",
        return_value=mock_get_activity_streams(streams_json),
    )

    # we don't use access token here, since we will mock the stravalib json response
    activity = read_strava(
        activity_id=4437021783,
        access_token=None,
        refresh_token=None,
        client=strava_client,
        to_df=False,
    )
    assert isinstance(activity, types.Activity)
    included_data = set(
        [
            "alt",
            "cad",
            "dist",
            "hr",
            "lon",
            "lat",
            "moving",
            "velocity_smooth",
            "grade_smooth",
        ]
    )
    assert included_data <= set(activity.columns.to_list())
    assert activity.size == 15723


test_data = [
    (pytest.lazy_fixture("strava_activity"), "alt", 0, 6.4),
    (pytest.lazy_fixture("strava_activity"), "alt", -1, 6.6),
    (pytest.lazy_fixture("strava_activity"), "cad", 0, 79),
    (pytest.lazy_fixture("strava_activity"), "cad", -1, 86),
    (pytest.lazy_fixture("strava_activity"), "dist", 0, 0.0),
    (pytest.lazy_fixture("strava_activity"), "dist", -1, 12019.7),
    (pytest.lazy_fixture("strava_activity"), "hr", 0, 111),
    (pytest.lazy_fixture("strava_activity"), "hr", -1, 160),
    (pytest.lazy_fixture("strava_activity"), "lat", 0, -8.016994),
    (pytest.lazy_fixture("strava_activity"), "lon", 0, -34.847439),
    (pytest.lazy_fixture("strava_activity"), "lat", -1, -8.016821),
    (pytest.lazy_fixture("strava_activity"), "lon", -1, -34.84716),
    (pytest.lazy_fixture("strava_activity"), "moving", 0, False),
    (pytest.lazy_fixture("strava_activity"), "moving", -1, True),
    (pytest.lazy_fixture("strava_activity"), "velocity_smooth", 0, 0.0),
    (pytest.lazy_fixture("strava_activity"), "velocity_smooth", -1, 3.2),
    (pytest.lazy_fixture("strava_activity"), "grade_smooth", 0, 1.1),
    (pytest.lazy_fixture("strava_activity"), "grade_smooth", -1, -0.6),
    (pytest.lazy_fixture("strava_dataframe"), "altitude", 0, 6.4),
    (pytest.lazy_fixture("strava_dataframe"), "altitude", -1, 6.6),
    (pytest.lazy_fixture("strava_dataframe"), "cadence", 0, 79),
    (pytest.lazy_fixture("strava_dataframe"), "cadence", -1, 86),
    (pytest.lazy_fixture("strava_dataframe"), "distance", 0, 0.0),
    (pytest.lazy_fixture("strava_dataframe"), "distance", -1, 12019.7),
    (pytest.lazy_fixture("strava_dataframe"), "heartrate", 0, 111),
    (pytest.lazy_fixture("strava_dataframe"), "heartrate", -1, 160),
    (pytest.lazy_fixture("strava_dataframe"), "latitude", 0, -8.016994),
    (pytest.lazy_fixture("strava_dataframe"), "longitude", 0, -34.847439),
    (pytest.lazy_fixture("strava_dataframe"), "latitude", -1, -8.016821),
    (pytest.lazy_fixture("strava_dataframe"), "longitude", -1, -34.84716),
    (pytest.lazy_fixture("strava_dataframe"), "moving", 0, False),
    (pytest.lazy_fixture("strava_dataframe"), "moving", -1, True),
    (pytest.lazy_fixture("strava_dataframe"), "velocity_smooth", 0, 0.0),
    (pytest.lazy_fixture("strava_dataframe"), "velocity_smooth", -1, 3.2),
    (pytest.lazy_fixture("strava_dataframe"), "grade_smooth", 0, 1.1),
    (pytest.lazy_fixture("strava_dataframe"), "grade_smooth", -1, -0.6),
]


@pytest.mark.parametrize("activity,column,index,expected", test_data)
def test_strava_values(activity, column, index, expected):
    assert activity[column].iloc[index] == expected
    assert activity.index[-1] == Timedelta("0 days 01:25:45")

    if isinstance(activity, types.Activity):
        assert activity.start == Timestamp("2020-12-06 06:36:27")



"""
Tests for utils module
"""
import os
import pytest
from pandas import Timedelta
from runpandas import _utils as utils

pytestmark = pytest.mark.stable


@pytest.fixture(scope="session")
def valid_tcx_filename(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("activity.tcx")


@pytest.fixture(scope="session")
def invalid_tcx_filename(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("activity.tcx.gz")


def test_file_exists(valid_tcx_filename):
    f = open(valid_tcx_filename, "w")
    f.write("test")
    f.close()  # open and close to toucj
    assert utils.file_exists(valid_tcx_filename)


def test_file_not_exists(invalid_tcx_filename):
    assert not utils.file_exists(invalid_tcx_filename)


def test_splitext_plus(valid_tcx_filename):
    prefix, sufix = utils.splitext_plus(valid_tcx_filename)
    assert os.path.basename(prefix) == "activity"
    assert sufix == ".tcx"


def test_splitext_plus_with_zip(invalid_tcx_filename):
    prefix, sufix = utils.splitext_plus(invalid_tcx_filename)
    assert os.path.basename(prefix) == "activity"
    assert sufix == ".tcx.gz"


def is_valid(valid_tcx_filename):
    assert utils.is_valid(valid_tcx_filename)
    assert not utils.is_valid("invalid_tcx_filename")


def is_valid_nok(invalid_tcx_filename):
    assert not utils.is_valid(invalid_tcx_filename)


def test_camecase_to_snakecase():
    assert utils.camelcase_to_snakecase("LatitudeDegrees") == "latitude_degrees"
    assert utils.camelcase_to_snakecase("Time") == "time"


def test_convert_pace_secmeters2minkms():
    assert utils.convert_pace_secmeters2minkms(
        Timedelta("0 days 00:00:00.407078").total_seconds()
    ) == Timedelta("0 days 00:06:47")
